# This Python file uses the following encoding: utf-8
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


DPAY_100_PERCENT = 10000
DPAY_1_PERCENT = 100
DPAY_REVERSE_AUCTION_WINDOW_SECONDS_HF20 = 900
DPAY_REVERSE_AUCTION_WINDOW_SECONDS_HF6 = 1800
DPAY_VOTE_REGENERATION_SECONDS = 432000
DPAY_VOTING_MANA_REGENERATION_SECONDS = 432000
DPAY_VOTE_DUST_THRESHOLD = 50000000
DPAY_ROOT_POST_PARENT = ''
