<!DOCTYPE html>
<html lang="en">
<head>
  <title>{{title}}</title>
  <meta charset="UTF-8">
  <meta name="description" content="{{description}}">
  <meta id="viewport" name="viewport" content="initial-scale=1.0, maximum-scale=1.0, user-scalable=0"/>
  <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
  {{baseCss}}
  <link rel="stylesheet" href="/css/ie.css" type="text/css" media="all">
  <!--[if lt IE 9]><script src="//html5shiv.googlecode.com/svn/trunk/html5.js"></script><![endif]-->
  <script src='//www.google.com/recaptcha/api.js'></script>
</head>
<body>
  <div id="app"></div>
  <script type="text/javascript">
    window.config = {{{clientConfig}}};
  </script>
  <script type="text/javascript" src="/js/app.min.js"></script>
</body>
</html>
